// add SAVI video player stylesheet
(function () {
    var video_head = document.getElementsByTagName('head')[0];
    var video_style = document.createElement('link');
    video_style.href = 'https://savi-cdn.macmillantech.com/fonts/SourceSansPro/sourceSansPro.css';
    video_style.type = 'text/css';
    video_style.rel = 'stylesheet';
    video_head.appendChild(video_style);
})();

// add custom SAVI CSS overrides to make responsive
(function () {
    var video_head = document.getElementsByTagName('head')[0];
    var video_style = document.createElement('link');
    video_style.href = 'asset/css/savi_video_local.css';
    video_style.type = 'text/css';
    video_style.rel = 'stylesheet';
    video_head.appendChild(video_style);
})();

$('[data-type="section"]').removeAttr('level');

$('[http-equiv="X-UA-Compatible"]').remove();

(function ($) {
    $.fn.changeElementType = function (newType) {
        var attrs = {};

        $.each(this[0].attributes, function (idx, attr) {
            attrs[attr.nodeName] = attr.nodeValue;
        });

        this.replaceWith(function () {
            return $("<" + newType + "/>", attrs).append($(this).contents());
        });
    };
})(jQuery);

$('br.clear').remove();

$('#manuscript').changeElementType('main');

$('div[data-type="question"]').each(function () {
    $(this).changeElementType('form');
});

$('[data-type="query_display"]').remove();

$('[data-type="query"]').each(function () {
    $(this).changeElementType('div');
});

$('[data-element]').each(function () {
    var $this = $(this);
    var el = $this.attr('data-element');
    $this.removeAttr('data-element');
    $this.changeElementType(el);
});

$('[data-class]').each(function () {
    var $this = $(this);
    var classes = $this.attr('data-class');
    $this.removeAttr('data-class');
    $this.addClass(classes);
});

$('h1').each(function () {
    var $this = $(this);
    var h1_html = $this.find('[data-type="title"]').html();
    $this.html(h1_html);
});

var Activity_subtype = Activity.extend({

    ARGA_initialization_alert: function () {
        player.activity.ARGA_initialization_final();
    },

    add_question: function (index, element) {
        var a = player.activity;	// this fn gets called within the scope of a jquery each, so we can't refer to "this"

        var jq = $(element);

        // see if we're in a question_pool or question_sequence block
        var qs_parent = jq.parents("[data-block_type=question_sequence], [data-block_type=question_sequence_one_at_a_time]");

        var q = a.questions[a.questions.length] = new Question({
            "index": a.questions.length
            , "number": jQuery.trim($("[data-type='number']", jq).html())
            , "title": jQuery.trim($("[data-type='title']", jq).html())
            , "jq": jq
            , "points_possible": 0
        });

        // store the question index in the DOM too
        jq.attr("data-question_index", q.index);

        // get question queries		
        a.question_being_initialized = q;
        $("[data-type='query']", jq).each(a.add_query);

        // if the question has queries, add bling to the bottom
        if (q.queries.length > 0) {


            var html = "<div class='question_error_div' id='question_error_div_" + q.index + "'></div>";

            html += "<div class='question_action_div' id='question_action_div_" + q.index + "'>";

            var submit_handler;
            // add a submit button if we're supposed to have one
            // (image map and matching questions, which should contain one and only one query, don't have submit buttons)
            if (q.queries[0].hasSubmitButton === true) {
                // "answer_one" pool mode (yet to be fully implemented...)
                if (q.pool_mode == 'answer_one') {
                    html += "<input class='standard_button question_submit_button' type='button' id='question_submit_button_" + q.index + "' value='Answer This Query' />";
                    submit_handler = player.activity.Show_Submit;

                } else {
                    html += a.question_submit_button(q);
                    submit_handler = player.activity.submit_question;
                }
            }

            // add a "give up" button if we're supposed to have one
            // (image map questions, which should contain one and only one query, should have this button)
            if (q.queries[0].hasGiveUpButton === true) {
                html += "<input class='standard_button question_submit_button' type='button' id='question_submit_button_" + q.index + "' value='Give Up' />";
                submit_handler = player.activity.give_up_question;
            }

            html += "</div>";

            jq.append(html);

            // bind the submit_handler function to the button, passing in the question_index
            $("#question_submit_button_" + q.index).click({ question_index: q.index }, submit_handler);

            // float instructor controls to right of action button
            jq.prepend(a.instructor_question_controls(q));

            // also, determine question points and weight at this time
            for (var i = 0; i < q.queries.length; ++i) {
                q.points_possible += q.queries[i].getPointsPossible();
            }
            // round to two decimal places, in case there are decimals
            q.points_possible = Math.round(q.points_possible * 100) / 100;
        }

        // now that the question has been created, deal with sequences and pools
        if (qs_parent.length > 0) {
            // get sequence index and item
            var qs_index = qs_parent.attr("question_sequence_index") * 1;
            var qs = a.question_sequences[qs_index];

            // place in sequence
            var sequence_num = qs.questions.length;
            qs.questions[sequence_num] = q;

            // place reference to sequence in question
            q.question_sequence = qs;
            q.question_sequence_index = sequence_num;

            // if this isn't #1, hide the question
            if (sequence_num > 0) {
                jq.hide();
            }

            // update the sequence navigator if it's there
            $("#question_sequence_max_index_" + qs.index).html(sequence_num + 1);
        }

        // Adjust heading
        var metadata = player.activity.questions[q.index].md;
        var $heading = $('[data-question_index="' + q.index + '"]').find('h3');
        if (metadata.question_number) {
            $heading.text('Question ' + metadata.question_number);
        }
        else {
            $heading.text('Question');
        }

        if (metadata.heading_hidden === 'true') {
            $heading.addClass('visually-hidden');
        }

        if (metadata.heading_level) {
            $heading.changeElementType('h' + metadata.heading_level);
        }

    },

    question_submit_button: function (q) {
        // define question_submit_button_label as a variable so it can be overridden
        var button_html = "<button class='standard_button question_submit_button' id='question_submit_button_"
            + q.index
            + "'>Submit<span class='visually-hidden'> Question "
            + q.index
            + "</span></button>";
        return button_html;
    },

    submit_question: function (question_index) {

        let event;

        if (typeof question_index == "object") {
            // question_index is actually the event object and we need 
            // save it off so we can stop propagation later on if
            // needed
            event = question_index;
            // extract the question_index from the object
            question_index = question_index.data.question_index;
        }

        var query = player.activity.questions[question_index].queries[0];

        if (query.getQueryTypeForARGA() === 'drop-down') {
            var userAnswer = $('#query_answer_' + query.query_index).val();
            if (Number(userAnswer) < 0) {
                var errMesg = player.activity.questions[question_index].md.input_error || 'Error!';
                query.setError(errMesg);
                AriaText.setLiveText(errMesg);
                return false;
            }
            else {
                query.clearError(errMesg);
            }
        }
        else if (query.getQueryTypeForARGA() === 'essay') {
            var userAnswer = $('#query_answer_' + query.query_index).val().trim();
            if (userAnswer === "") {
                var errMesg = "You must enter an answer before submitting."
                query.setError(errMesg);
                AriaText.setLiveText(errMesg);
                return false;
            }
            else {
                query.clearError(errMesg);
            }
        }

        event.preventDefault();
        event.stopPropagation();

        if (query.getQueryTypeForARGA() === 'essay') {

            ARIAmodal.openModal(
                {
                    modalId: 'aria_modal',
                    top: player.getModalTop(),
                    title: 'Verify Submission',
                    text: 'Do you want to submit your answer? You will not be able to edit your answer after submitting.',
                    buttons: [
                        {
                            text: 'Cancel', onClick: function () {
                                player.setFocus('question_submit_button_' + question_index);
                            }
                        },
                        {
                            text: 'Submit', onClick: function () {
                                player.activity.submit_question_go(question_index);
                            }
                        }
                    ],
                    closeFocus: 'question_submit_button_' + question_index
                }
            );

            return;

        }
        else {
            player.activity.submit_question_go(question_index);
        }
    },

    submit_question_go: function (question_index) {

        // the arg may be a number or an event object; in the latter case...
        if (typeof question_index == "object") {
            // extract the question_index from the object
            question_index = question_index.data.question_index;
        }

        player.LAST_QUESTION_ANSWERED = question_index;

        var t = player.activity;

        var question = t.questions[question_index];

        var all_answers_correct = true;
        // for each query in the question...
        for (var i = 0; i < question.queries.length; ++i) {
            var query = question.queries[i];

            // evaluate the answer. If evaluateAnswer returns false, the answer isn't submittable at this time -- 
            // most likely because it's a free-response answer that needs to be evaluated via MAVE.
            var answer_submittable = query.evaluateAnswer(question_index);
            if (answer_submittable === false) {
                // as soon as we get an unsubmittable answer, we return. submit_question
                // will be called again, either by the callback fn of MAVE or because the student
                // resubmits. At that point we'll catch any additional free-response answers that need
                // checking, or we'll finish this loop here and move on.
                return;
            }
            all_answers_correct = all_answers_correct && query.isCorrect();
        }

        // if we made it to here, we're at least going to show which items are correct and incorrect
        // we might overwrite with review mode below, but that's OK
        for (var i = 0; i < question.queries.length; ++i) {
            var query = question.queries[i];
            query.jq.html(query.getHTML("review_correct_incorrect"));
        }

        // update question tries
        ++question.tries;

        // if answer is incorrect, and max_tries is set, and # of tries is less than max_tries
        // (and the user didn't just give up)
        if ((!all_answers_correct && question.md.max_tries != null && question.tries < question.md.max_tries) && question.gave_up != true) {
            // tell user he can try again
            t.show_question_feedback(question, false, true);
            return;
        } else {
            t.show_question_feedback(question, all_answers_correct, false);
        }

        //$('#question_feedback_' + question_index).focus();

        // if we make it to here we need to set ARGA info for the question, and show queries in review mode
        // if md.question_report_text is set, we submit the question as a whole, rather than as individual queries
        var query_correct_answers = "";
        var query_responses = "";
        var query_responses_for_questionData = "";
        var query_grade_sum = 0;
        // for each query in the question...
        for (var i = 0; i < question.queries.length; ++i) {
            var query = question.queries[i];

            // establish a question number to use
            var query_number = t.get_query_number(question, i);

            // get a query weight
            var query_weight = question.get_weight();
            // if the question has more than one query, divide weight accordingly
            if (question.queries.length > 1) {
                query_weight = Math.round(query_weight / question.queries.length);
            }

            // if credit_after_max_tries is set, give at least that credit now, as long as the student tried
            if (question.md.credit_after_max_tries != null) {
                var query_grade = query.getGradeForARGA();
                if (query_grade >= 0 && query_grade < question.md.credit_after_max_tries) {
                    query.overrideGrade(question.md.credit_after_max_tries * 1);
                }
            }

            // set the ARGA response, unless we're saving one response for the whole question
            if (!question.md_set("question_report_text")) {
                t.ARGA_set_query_response(query, query_number, query_weight);

                // otherwise compose query roll-ups
            } else {
                if (query_correct_answers != "") {
                    query_correct_answers += " / ";
                    query_responses += " / ";
                    query_responses_for_questionData += t.query_responses_separator;
                }
                query_correct_answers += query.getCorrectAnswerForARGA();
                query_responses += query.getUserAnswerForARGA();
                // see restore_saved_answers for how this is used later
                query_responses_for_questionData += query.getUserAnswerForARGA()
                    + t.query_responses_grade_separator
                    + query.getOriginalGrade();
                query_grade_sum += query.getGradeForARGA() * 1;
            }

            // show in review mode, unless question metadata says not to show correct answers
            if (question.md.show_correct_answers != "false") {
                query.jq.html(query.getHTML("review"));
            }
        }

        // if we're saving one response for the whole question, save here
        if (question.md_set("question_report_text")) {
            var question_grade = Math.round(query_grade_sum / question.queries.length);
            t.ARGA_set_question_response(question, query_correct_answers, query_responses, question_grade, query_responses_for_questionData);
        }

        // grade the activity
        t.grade_activity();

        // then if ARGA is running, submit data; show_grade_saved_message will be called on completion.
        if (player.ARGA_running) {
            t.ARGA_submit_data();
        }

        if (this.allQuestionsAnswered()) {

            $('#completed-mesg').attr('tabindex', '-1').insertAfter($('#wrapper')).addClass('completed').attr('aria-labelledby', 'completed_heading completed_content').removeAttr('hidden');

            // If we are in PX then this will be called 
            // in show_grade_saved_message.
            if (!player.viewingInPx()) {
                setTimeout(function () {
                    $('#completed-mesg').focus();
                }, 500);
            }
        }
        else {
            // focus on feedback
            setTimeout(function () {
                $('#question_feedback_' + question_index).focus();
            }, 500);
        }

        // update the question_action_div
        t.update_question_action_div(question_index);

        // if the item is in a sequence, show the next question
        t.show_next_question(question);

        // update section status
        player.update_section_status();

        var s = player.get_current_section();

        // if all questions are answered and all sequences are complete
        // (these are set by _super)
        if (s.all_questions_answered && s.all_question_sequences_completed) {
            // if we're not on the final step
            if (s.index < player.sections.length - 1) {
                // then show the buttona
                //UI_Elements.toggle_button("step_button", true);
                $('#step_link').attr('aria-disabled', 'false');

                // otherwise we're done!
            }
            else {
                //player.show_completion_message();
            }
        }

        player.update_navigation();

    },

    show_question_feedback: function (question, is_correct, try_again) {
        // remove prior feedback if there
        $("#question_feedback_" + question.index).remove();

        // if we don't have any feedback at all, return
        /*
        if (question.md.feedback == null && question.md.feedback_correct == null && question.md.feedback_incorrect == null && question.md.feedback_hint == null) {
            return;
        }
        */

        var fbHeadingLevel = question.md.heading_level ? (Number(question.md.heading_level) + 1) : 4;

        var html = "<div class='question_feedback' id='question_feedback_" + question.index + "' tabindex='-1'>"
            + "<h" + fbHeadingLevel + " class='question_feedback_head'>"
            + "<span class='visually-hidden'>Question "
            + question.index + " </span>"
            + this.feedbackTitle + "</h" + fbHeadingLevel + ">";

        var sr_fb = question.md.feedback_sr ? " <span class='visually-hidden'>" + question.md.feedback_sr + "</span>" : "";

        if (player.md.submit_response) {
            html += "<div class='question_feedback_part'>"
                + player.md.submit_response
                + "</div>";
        }

        // include general feedback if there
        if (question.md.feedback != null) {
            html += "<div class='question_feedback_part'>" + question.md.feedback + sr_fb + "</div>";
        }

        // include correct feedback if there and the answer is correct
        if (is_correct && question.md.feedback_correct != null) {
            html += "<div class='question_feedback_part'>" + question.md.feedback_correct + "</div>";
        }

        // close question_feedback div
        html += "</div>";

        // append the div, then fade it in
        $("#question_action_div_" + question.index).prepend(html);
        $("#question_feedback_" + question.index).fadeIn("fast", function () {
            // What I'd like to do here is check to make sure the feedback is all showing on the screen,
            // and scroll if necessary so that the feedback is all visible. But just doing "scrollTo"
            // isn't working properly in all situations.
            // $(window).scrollTo($(this), 0);
        });
    },

    update_question_action_div: function (question_index) {
        var question = this.questions[question_index];

        // calculate points_awarded and points_possible
        var points_awarded = 0;
        var points_possible = 0;
        var points_provisional = false;
        for (var i = 0; i < question.queries.length; ++i) {
            var query = question.queries[i];
            points_awarded += query.getPointsAwarded();
            points_possible += query.getPointsPossible();
            if (query.pointsAreProvisional() == true) {
                points_provisional = true;
            }
        }

        // round to at most two decimal places
        points_awarded = Math.round(points_awarded * 100) / 100;
        points_possible = Math.round(points_possible * 100) / 100;

        // hide submit button (and any other similar buttons that might be added by subtypes), but show points earned
        var jq = $("#question_action_div_" + question.index);
        //jq.find(".question_submit_button").hide();
        jq.find(".question_submit_button").remove();
        var points_html = "<div class='points_awarded_div'>Points awarded: " + points_awarded + " out of " + points_possible;

        if (points_provisional) {
            points_html += " (<a href='javascript:player.activity.explain_provisional()'>provisional</a>)";
        }
        points_html += "</div>";

        var jq = $("#question_action_div_" + question.index);
        jq.find(".points_awarded_div").remove();
        /*
        if (jq.find(".question_feedback").length > 0) {
            jq.find(".question_feedback").after(points_html);
        }
        else {
            jq.prepend(points_html);
        }
        */
    },

    allQuestionsAnswered: function () {

        var all_questions_answered = true;
        $("[data-type=question]").each(function (index, element) {
            var q_index = $(element).attr("data-question_index");
            var question = player.activity.questions[q_index];

            all_questions_answered = all_questions_answered && question.is_complete();
        });
        return all_questions_answered;
    },

    show_grade_saved_message: function () {
        // if the user has answered all queries AND the user hasn't been told
        // this already, tell them now
        if (player.activity.answered_all_queries && !player.user_has_received_completion_alert) {
            //setTimeout('Standard_Dialog.alert("You have completed the activity, and your grade has been submitted.");', 20);
            player.user_has_received_completion_alert = true;
        }

        // brb: This method gets called by Save_ARGA_Data_Callback so we
        // can add anything here that we want to have happen after the
        // activity has been completed.
        if (player.activity.allQuestionsAnswered()) {
            setTimeout(function () {
                $('#completed-mesg').focus();
            }, 500);
        }
    },

});

var Player_subtype = Player.extend({

    ACTIVITY_COMPLETED: false,
    DUE_DATE_HAS_PASSED: false,

    viewingInPx: function () {
        if (typeof window.PX !== 'undefined') {
            return true;
        }
        if (typeof window.parent.PX !== 'undefined') {
            return true;
        }
        return false;
    },

    setFocus: function (arg) {
        if (typeof arg === 'string') {
            arg = document.getElementById(arg);
        }
        setTimeout(function () {
            try {
                arg.focus();
            }
            catch (err) {
                console.log('setFocus: ' + err.message);
            }
        }, 250);
    },

    getModalTop: function () {
        var fneContent = window.parent.document.getElementById('fne-content');
        if (fneContent === null || fneContent === undefined) {
            return undefined;
        }
        var fneHeight = fneContent.clientHeight;
        var scrollTop = fneContent.scrollTop;
        return scrollTop + (fneHeight / 2);
    },

    initialize: function () {

        // for scrolling issues
        if (player.viewingInPx()) {
            $('html').addClass('px');
        }

        this._super();

        this.activity = new Activity_subtype();
    },

    initialize2: function () {
        this._super();

        /*  Load savi brightcove library.
            We have to load this here because there is an issue in 
            Variables.init in which the entire html in the manuscript div is
            replaced and we have to wait for that to happen before activiating 
            the savi-brightcove library. */
        $.getScript("https://savi-cdn.macmillantech.com/brightcove/savi-brightcove-player.js", function (data, textStatus, jqxhr) { });

        /* Remove unused metadata just to clean things up. */
        $('[data-type="metadata"]').remove();

        /* Fix-up h3 Question headings. */
        $('form > h3').each(function () {
            var $this = $(this);
            var qNum = $this.parent().attr('data-question_index');
            $this.html('<span class="visually-hidden">Question </span>' + qNum);
        });

        // we need to do this again to catch these in questions stems
        $('[data-element]').each(function () {
            var $this = $(this);
            var el = $this.attr('data-element');
            $this.removeAttr('data-element');
            $this.changeElementType(el);
        });

        /*
        tinymce.init({
            selector: '#query_answer_0',
            statusbar: false,
            toolbar: false,
            menubar: 'edit',
            width: '100%',
            height: '20rem',
        });
        */

        if (player.ARGA_running) {
            // brb: determine if activity has been completed (all ARGA answered)
            try {
                var dejs_arr = ARGA_API.dejs_data;
                for (var i = 0; i < dejs_arr.length; i++) {
                    if (dejs_arr[i].name === "cmi.completion_status") {
                        this.ACTIVITY_COMPLETED = (ARGA_API.dejs_data[i].value == "completed");
                        break;
                    }
                }
            }
            catch (err) {
                safe_log("initialize2: Can't get cmi.completion_status: " + err)
            }

            // brb: determine if due date has passed for assignment
            if (player.ARGA_running) {
                if (Get_ARGA_Data("due_date_has_passed") == 1) {
                    this.DUE_DATE_HAS_PASSED = true;
                }
            }
        }
        else {
            try {
                const params = new URLSearchParams(window.location.search);
                if (params.has('pastdue') && params.get('pastdue') === 'true') {
                    this.DUE_DATE_HAS_PASSED = true;
                }
            }
            catch (e) { }
        }

        if (this.DUE_DATE_HAS_PASSED && !this.ACTIVITY_COMPLETED) {

            $('button.question_submit_button').each(function () {
                var $this = $(this);
                $this.remove();
            });

            // disable all questions
            $('[data-type="question"]').each(function () {
                var $this = $(this);
                $this.find('textarea').attr("aria-disabled", true).attr('readonly', true).attr('tabindex', '-1');
            });

            $('#close_expired').click(function () {
                $('h1').focus();
                $('#expired-mesg').remove();
            });

            $('#completed-mesg').remove();
            $('#expired-mesg').removeAttr('hidden');
            return;
        }

        if (!this.DUE_DATE_HAS_PASSED) {
            $('#expired-mesg').remove();
        }

        if (!this.ACTIVITY_COMPLETED) {
            // brb: We don't want to remove completed message because we
            // are going to show it when they do complete activity.
            //$('#completed-mesg').remove();
        }
        else {
            $('#completed-mesg').removeAttr('hidden');

            $('#close_completed').click(function () {
                $('h1').focus();
                $('#completed-mesg').remove();
            });
        }

        AriaText.initLiveFeedback();
    },

    /* brb: Overriding this method in order to keep the activity from 
       updating ARGA if the activity has been completed or hasn't and is passed due date
       so that it doesn't overwrite a manual grade change in the grade book (see JIRA LP-127) */
    update_section_status: function () {
        if (this.section_currently_showing != null && this.last_available_section <= this.section_currently_showing) {
            var s = this.get_current_section();
            s.all_questions_answered = true;

            // if the current section has any questions, see if they've all been answered
            var qs_to_check = [];
            s.jq.find("[data-type=question]").each(function (index, element) {
                var q_index = $(element).attr("data-question_index");
                var question = player.activity.questions[q_index];
                s.all_questions_answered = s.all_questions_answered && question.is_complete();
                if (question.question_sequence != null) {
                    for (var i = 0; i < qs_to_check.length; ++i) {
                        if (qs_to_check[i] == question.question_sequence) break;
                    }
                    if (i == qs_to_check.length) {
                        qs_to_check.push(question.question_sequence);
                    }
                }
            });

            // now check question sequences to make sure they're done too
            s.all_question_sequences_completed = true;
            for (var i = 0; i < qs_to_check.length; ++i) {
                s.all_question_sequences_completed = s.all_question_sequences_completed && (qs_to_check[i].sequence_complete == "complete");
            }

            if (s.all_questions_answered && s.all_question_sequences_completed) {
                this.last_available_section = this.section_currently_showing + 1;
            }
        }

        this.fill_in_previous_query_responses();

        // if ARGA is initialized and we need to restore the user to his last position when he returns,
        if (this.ARGA_running && player.md.restore_last_viewed_section == "true") {
            // if we're viewing a page other than the one currently saved as LVS (last-viewed section)
            var lvs = Get_ARGA_Data("LVSX");
            if (this.section_currently_showing != null && this.section_currently_showing != lvs) {
                // store LVS/LAS to register the last-viewed/last-available page,
                // and save ARGA so it gets registered
                Set_ARGA_Data("LVSX", this.section_currently_showing);
                Set_ARGA_Data("LASX", this.last_available_section);

                /* brb: If the activity has not been completed then save LVSX, but if
                  it has been completed then do not save LVSX because that will also
                  send the grade to the PX grade book again, overwriting any manual
                  grade changes the instructor may have made. */
                /* brb: also don't want to do this if the due date has passed */
                if (!this.ACTIVITY_COMPLETED && !this.DUE_DATE_HAS_PASSED) {
                    this.activity.ARGA_submit_data(false);
                }
            }
        }
    },

    show_section: function (section_to_show) {
        this._super(section_to_show);
        setTimeout(function () {
            try {
                parent.PX.API.Content.ResetHeight();
            }
            catch (e) { }
        }, 500);
    }

});

player = new Player_subtype();

Essay_Query = Essay_Query.extend({

    userHasAnswered: function () {
        //return (this.user_answer != "" && this.user_answer != null);
        return (this.grade == 100);
    },

    query_text_html: function () {
        var prefix = '';
        if (this.parent_question.index) {
            prefix = '<span class="visually-hidden">Question ' + this.parent_question.index + ': </span>';
        }

        if (this.query_text != "" && this.md._hide_query_text != "true") {
            // replace newlines with <br />
            var qt = jQuery.trim(this.query_text).replace(/\n/g, "<br>")

            return "<label class='query_text' for='query_answer_" + this.query_index + "'>" + prefix + qt + "</label>"
        } else {
            return "";
        }
    },

    getHTML: function (mode) {
        // start with the query text, if there
        var html = this.query_text_html();

        var val, disabled;
        if (!this.userHasAnswered() || mode == 'preview') {
            val = "";
            disabled = "";

            // otherwise we have an answer, so fill it in
        } else {
            // make sure we convert double quotes to &#34; for placement in 'value="xxx"' attributes
            val = this.user_answer;

            // convert breaks to newlines
            val = val.replace(/<br *(\/)?>/g, "\n");

            // also disable the input if we're not in review_correct_incorrect mode
            if (mode != "review_correct_incorrect") {
                disabled = "aria-disabled='true' readonly tabindex='-1'";
            }
        }

        html += "<textarea id='query_answer_" + this.query_index + "' class='query_essay_ta' " + disabled + ">"
            + val + "</textarea>";

        // if answer was just entered and grade is 100, also show the "provisional" message
        // (unless for this question it isn't provisional)
        /*
        if (this.answer_was_just_entered && this.grade == 100 && this.md.points_are_provisional == "true") {
            html += "<div class='essay_question_evaulation_feedback'>Your answer has been provisionally accepted. You'll get full credit for now, but your instructor may update your grade later after evaluating it.</div>";
        }
        */

        return html;
    },

});

Query.prototype.setError = function (errMesg) {

    var alertIcon = '<svg version="1.0" xmlns="http://www.w3.org/2000/svg" width="1280.000000pt" height="1138.000000pt" viewBox = "0 0 1280.000000 1138.000000" preserveAspectRatio="xMidYMid meet" role="image" aria-hidden="true"><g transform="translate(0.000000,1138.000000) scale(0.100000,-0.100000)" stroke="none"><path d="M6280 11374 c-14 -2 -51 -9 -83 -15 -252 -45 -510 -210 -661 -422 -24 -34 -754 -1293 -1621 -2797 -868 -1504 -2077 -3601 -2688 -4660 -611 -1059 -1125 -1959 -1142 -2000 -119 -280 -110 -601 25 -876 140 -287 381 -485 698 -572 l97 -27 5495 0 5495 0 97 27 c317 87 558 285 698 572 136 277 145 590 25 876 -17 41 -531 941 -1142 2000 -611 1059 -1820 3156 -2688 4660 -867 1504 -1597 2763 -1621 2797 -189 266 -496 428 -829 438 -71 2 -141 2 -155 -1z m137 -1086 c6 -13 330 -576 720 -1253 390 -676 1321 -2290 2068 -3585 747 -1295 1621 -2809 1941 -3365 l583 -1010 -2665 -3 c-1465 -1 -3863 -1 -5328 0 l-2665 3 289 500 c1005 1744 2443 4238 3645 6322 765 1327 1393 2412 1396 2413 3 0 10 -10 16 -22z"/><path d="M6275 7456 c-338 -64 -584 -357 -585 -694 0 -45 27 -233 70 -490 257 -1521 290 -1713 300 -1750 23 -85 111 -185 198 -225 67 -31 217 -31 284 0 87 40 175 140 198 225 5 18 39 209 75 423 36 215 86 507 110 650 187 1098 188 1105 182 1205 -14 258 -166 484 -402 596 -134 64 -294 86 -430 60z"/><path d="M6313 3549 c-415 -48 -701 -459 -603 -868 62 -257 270 -466 525 -527 90 -21 240 -21 330 0 377 90 615 478 525 855 -84 351 -416 581 -777 540z"/></g></svg>';

    $('#question_error_div_' + this.parent_question.index).html(alertIcon + errMesg).css('display', 'block');
}

Query.prototype.clearError = function () {
    $('#question_error_div_' + this.parent_question.index).html('').css('display', 'none');
}

var AriaText = {

    setLiveText: function (text) {
        var $liveTextDiv = $('#live-feedback');
        var currTime = Date.now();
        $liveTextDiv.attr('data-lasttime', currTime);
        $liveTextDiv.html('');
        $liveTextDiv.html(text);

        setTimeout(function () {
            var lastTime = $liveTextDiv.attr('data-lasttime');
            if (Number(lastTime) === currTime) {
                $liveTextDiv.html('');
            }
        }, 3000);

    },

    initLiveFeedback: function () {
        var liveFeedbackDiv = $('<div id="live-feedback" class="visually-hidden" role="alert"></div>');
        $('body').append(liveFeedbackDiv);
    }

};
